# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) 2015 European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Cif parser helper functions"""

__author__ = "Jérôme Kieffer"
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "2014, European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "27/10/2025"

import numpy
import cython


@cython.boundscheck(False)
def split_tokens(bytes_text):
    """
    Separate the text representing a CIF file into a list of tokens.

    :param bytes_text: the content of the CIF - file
    :type bytes_text:  8-bit string (str in python2 or bytes in python3)
    :return: list of all the fields of the CIF
    :rtype: list
    """
    cdef:
        unsigned char[:] ary = bytearray(bytes_text)
        bint in_comment = False, in_single_quote = False
        bint in_double_quote = False, multiline = False, go_on = True
        int i = -1, start = -1, end = -1, imax
        char prev, next, cur = b"\n"
        bytes EOL = b'\r\n'
        bytes BLANK = b" \t\r\n"
        unsigned char SINGLE_QUOTE = b"'"
        unsigned char DOUBLE_QUOTE = b'"'
        unsigned char SEMICOLUMN = b';'
        unsigned char HASH = b"#"
        unsigned char UNDERSCORE = b"_"
        unsigned char DASH = b"-"
        unsigned char QUESTIONMARK = b"?"
        bytes BINARY_MARKER = b"--CIF-BINARY-FORMAT-SECTION--"
        int lbms = len(BINARY_MARKER)
    next = ary[0]
    imax = len(bytes_text) - 1
    fields = []
    while go_on:
        i += 1
        prev = cur
        cur = next
        if i < imax:
            next = ary[i + 1]
        else:
            next = b"\n"
            go_on = False
        # print(i,chr(prev),chr(cur),chr(next),in_comment,in_single_quote,in_double_quote,multiline, start, cur ==SINGLE_QUOTE)
        # Skip comments
        if in_comment:
            if cur in EOL:
                in_comment = False
            continue

        if prev in EOL:
            if cur == HASH:
                in_comment = True
                continue
            if cur == SEMICOLUMN:
                if multiline:
                    fields.append(bytes_text[start:i].strip())
                    start = -1
                    multiline = False
                else:
                    multiline = True
                    start = i + 1
                continue

        if multiline:
            # Handle CBF
            if cur == DASH:
                if bytes_text[i:i + lbms] == BINARY_MARKER:
                    end = bytes_text[i + lbms:].find(BINARY_MARKER)
                    i += end + 2 * lbms
                    cur = ary[i]
                    next = ary[i + 1]
            continue

        # Handle single quote
        if cur == SINGLE_QUOTE:
            if (not in_single_quote) and (not in_double_quote) and (start < 0) and (prev in BLANK):
                start = i + 1
                in_single_quote = True
                continue
            if (in_single_quote) and (not in_double_quote) and (start >= 0) and (next in BLANK):
                fields.append(bytes_text[start:i].strip())
                start = -1
                in_single_quote = False
                continue
        if in_single_quote:
            continue

        # Handle double quote
        if cur == DOUBLE_QUOTE:
            if (not in_single_quote) and (not in_double_quote) and (start < 0) and (prev in BLANK):
                start = i + 1
                in_double_quote = True
                continue
            if (not in_single_quote) and (in_double_quote) and (start >= 0) and (next in BLANK):
                fields.append(bytes_text[start:i].strip())
                start = -1
                in_double_quote = False
                continue
        if in_double_quote:
            continue

        # Normal fields
        if cur in BLANK:
            if start >= 0:
                fields.append(bytes_text[start:i].strip())
                start = -1
        else:
            if start < 0:
                start = i
    if start >= 0:
        fields.append(bytes_text[start:].strip())
    return fields



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

"""Test suite for all fabio modules."""

import sys
import logging
import unittest
from . import test_fabio_image
from . import test_filenames
from . import test_file_series
from . import test_filename_steps
from . import test_header_not_singleton
from . import test_open_header
from . import test_open_image
from . import test_flat_binary
from . import test_compression
from . import test_nexus
from . import test_fabio_convert
from . import test_failing_files
from . import test_formats
from . import test_image_convert
from . import test_tiffio
from . import test_frames
from . import test_fabio
from . import codecs
from . import test_agi_bitfield
from . import test_densification
from . import test_io_limits
from . import test_utils_cli
from . import test_import

logger = logging.getLogger(__name__)


def suite():
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_fabio_image.suite())
    testSuite.addTest(test_filenames.suite())
    testSuite.addTest(test_file_series.suite())
    testSuite.addTest(test_filename_steps.suite())
    testSuite.addTest(test_header_not_singleton.suite())
    testSuite.addTest(test_open_header.suite())
    testSuite.addTest(test_open_image.suite())
    testSuite.addTest(test_flat_binary.suite())
    testSuite.addTest(test_compression.suite())
    testSuite.addTest(test_nexus.suite())
    testSuite.addTest(test_fabio_convert.suite())
    testSuite.addTest(test_failing_files.suite())
    testSuite.addTest(test_formats.suite())
    testSuite.addTest(test_image_convert.suite())
    testSuite.addTest(test_tiffio.suite())
    testSuite.addTest(test_frames.suite())
    testSuite.addTest(test_fabio.suite())
    testSuite.addTest(codecs.suite())
    testSuite.addTest(test_agi_bitfield.suite())
    testSuite.addTest(test_densification.suite())
    testSuite.addTest(test_io_limits.suite())
    testSuite.addTest(test_utils_cli.suite())
    testSuite.addTest(test_import.suite())
    return testSuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    if not runner.run(suite()).wasSuccessful():
        sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

"""Test failing files"""

import unittest
import os
import io
import fabio
import shutil
from .utilstest import UtilsTest


class TestFailingFiles(unittest.TestCase):
    """Test failing files"""

    @classmethod
    def setUpClass(cls):
        cls.tmp_directory = os.path.join(UtilsTest.tempdir, cls.__name__)
        os.makedirs(cls.tmp_directory)
        cls.createResources(cls.tmp_directory)

    @classmethod
    def createResources(cls, directory):
        cls.txt_filename = os.path.join(directory, "test.txt")
        with io.open(cls.txt_filename, "w+t") as f:
            f.write("Kikoo")

        cls.bad_edf_filename = os.path.join(directory, "bad_edf.edf")
        with io.open(cls.bad_edf_filename, "w+b") as f:
            f.write(b"\r{")
            f.write(b"\x00\xff\x99" * 10)

        cls.bad_edf2_filename = os.path.join(directory, "bad_edf2.edf")
        with io.open(cls.bad_edf2_filename, "w+b") as f:
            f.write(b"\n{\n\n}\n")
            f.write(b"\xff\x00\x99" * 10)

        cls.bad_msk_filename = os.path.join(directory, "bad_msk.msk")
        with io.open(cls.bad_msk_filename, "w+b") as f:
            f.write(b"M\x00\x00\x00A\x00\x00\x00S\x00\x00\x00K\x00\x00\x00")
            f.write(b"\x00\xff\x99" * 10)

        cls.bad_dm3_filename = os.path.join(directory, "bad_dm3.dm3")
        with io.open(cls.bad_dm3_filename, "w+b") as f:
            f.write(b"\x00\x00\x00\x03")
            f.write(b"\x00\xff\x99" * 10)

        cls.bad_npy_filename = os.path.join(directory, "bad_numpy.npy")
        with io.open(cls.bad_npy_filename, "w+b") as f:
            f.write(b"\x93NUMPY")
            f.write(b"\x00\xff\x99" * 10)

        cls.missing_filename = os.path.join(directory, "test.missing")

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmp_directory)

    def test_missing_file(self):
        self.assertRaises(IOError, fabio.open, self.missing_filename)

    def test_wrong_format(self):
        self.assertRaises(IOError, fabio.open, self.txt_filename)

    def test_wrong_edf(self):
        self.assertRaises(IOError, fabio.open, self.bad_edf_filename)

    def test_wrong_edf2(self):
        self.assertRaises(IOError, fabio.open, self.bad_edf_filename)

    def test_wrong_msk(self):
        self.assertRaises(ValueError, fabio.open, self.bad_msk_filename)

    def test_wrong_dm3(self):
        self.assertRaises(ValueError, fabio.open, self.bad_dm3_filename)

    def test_wrong_numpy(self):
        self.assertRaises(ValueError, fabio.open, self.bad_npy_filename)


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestFailingFiles))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fast Azimuthal integration
#             https://github.com/silx-kit/pyFAI
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Authors: Gael Goret <gael.goret@esrf.fr>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
"""
Portable diffraction images viewer/converter

* Written in Python, it combines the functionalities of the I/O library fabIO
  with a user friendly Qt GUI.
* Image converter is also a light viewer based on the visualization tool
  provided by the module matplotlib.
"""

__version__ = "1.1"
__author__ = "Gaël Goret, Jérôme Kieffer"
__copyright__ = "2015-2026 ESRF"
__licence__ = "MIT"
__date__ = "12/03/2026"


import sys
from argparse import ArgumentParser
import numpy
import fabio

# ----------------------------------------------------------------------
# Qt imports via QtPy – this works with PyQt5, PySide2, PySide6, etc.
# ----------------------------------------------------------------------
from qtpy.QtWidgets import QApplication, QStyleFactory
from ..qt.viewer import AppForm

# Matplotlib imports (unchanged)
# ----------------------------------------------------------------------
# Global configuration
# ----------------------------------------------------------------------
numpy.seterr(divide="ignore")

output_format = [
    "*.bin",
    "*.cbf",
    "*.edf",
    "*.h5",
    "*.img",
    "*.mar2300",
    "*.mar3450",
    "*.marccd",
    "*.tiff",
    "*.sfrm",
]


# ----------------------------------------------------------------------
# Entry point
# ----------------------------------------------------------------------
def main():
    parser = ArgumentParser(
        prog="fabio_viewer",
        usage="fabio_viewer img1 img2... imgn",
        description=__doc__,
        epilog=f"Based on FabIO version {fabio.version}",
    )
    parser.add_argument("images", nargs="*")
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=__version__,
        help="Print version & quit",
    )
    args = parser.parse_args()
    QApplication.setStyle(QStyleFactory.create("Cleanlooks"))
    app = QApplication([])
    form = AppForm()
    if args.images:
        form.open_data_series(args.images)
    form.show()
    return app.exec_()


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
"""
Deep test to check IOError exceptions
"""

import unittest
import os
import logging
from .utilstest import UtilsTest
import fabio

logger = logging.getLogger(__name__)


class TestImageConvert(unittest.TestCase):
    """Test image convertions"""

    def test_edf_to_tiff(self):
        tmpdir = os.path.join(UtilsTest.tempdir, self.id())
        os.mkdir(tmpdir)

        filename = UtilsTest.getimage("face.edf")
        output_filename = os.path.join(tmpdir, "face.tif")

        image = fabio.open(filename)
        image2 = image.convert("tiff")
        image2.save(output_filename)
        self.assertEqual(image.shape, image2.shape)
        image3 = fabio.open(output_filename)
        self.assertEqual(image.shape, image3.shape)


def suite():
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loader(TestImageConvert))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) 2015 European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Cif Binary Files images are 2D images written by the Pilatus detector and others.
They use a modified (simplified) byte-offset algorithm.  This file contains the
decompression function from a string to an int64 numpy array.
"""

__author__ = "Jérôme Kieffer"
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "2010-2016, European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "27/10/2025"


import numpy
import cython

from libc.stdint cimport int8_t, uint8_t, \
                         uint16_t, int16_t,\
                         int32_t, uint32_t,\
                         int64_t, uint64_t


@cython.boundscheck(False)
@cython.wraparound(False)
def comp_cbf32(data not None):
    """Compress a dataset using the byte-offset described for Pilatus

    :param data: array of integers
    :return: numpy array of chars
    """
    cdef:
        int32_t[::1] ary = numpy.ascontiguousarray(data.ravel(), dtype=numpy.int32)
        int size = ary.size, i = 0, j = 0
        int8_t[::1] output = numpy.zeros(size * 7, dtype=numpy.int8)
        int32_t last, current, delta, absdelta
    last = 0
    for i in range(size):
        current = ary[i]
        delta = current - last
        absdelta = delta if delta > 0 else -delta
        if absdelta >= 1 << 15:
            output[j] = -128
            output[j + 1] = 0
            output[j + 2] = -128
            output[j + 3] = (delta & 255)
            output[j + 4] = (delta >> 8) & 255
            output[j + 5] = (delta >> 16) & 255
            output[j + 6] = (delta >> 24)
            j += 7
        elif absdelta >= 1 << 7:
            output[j] = -128
            output[j + 1] = delta & 255
            output[j + 2] = (delta >> 8) & 255
            j += 3
        else:
            output[j] = delta
            j += 1
        last = current
    return numpy.asarray(output)[:j]


@cython.boundscheck(False)
@cython.wraparound(False)
def comp_cbf(data not None):
    """Compress a dataset using the byte-offset described for any int64

    :param data: array of integers
    :return: numpy array of chars
    """
    cdef:
        int64_t[::1] ary = numpy.ascontiguousarray(data.ravel(), dtype=numpy.int64)
        int size = ary.size, i = 0, j = 0
        int8_t[::1] output = numpy.zeros(size * 15, dtype=numpy.int8)
        int64_t last, current, delta, absdelta
    last = 0
    for i in range(size):
        current = ary[i]
        delta = current - last
        absdelta = delta if delta > 0 else -delta
        if absdelta >= 1 << 31:
            output[j] = -128
            output[j + 1] = 0
            output[j + 2] = -128
            output[j + 3] = 0
            output[j + 4] = 0
            output[j + 5] = 0
            output[j + 6] = -128
            output[j + 7] = (delta & 255)
            output[j + 8] = (delta >> 8) & 255
            output[j + 9] = (delta >> 16) & 255
            output[j + 10] = (delta >> 24) & 255
            output[j + 11] = (delta >> 32) & 255
            output[j + 12] = (delta >> 40) & 255
            output[j + 13] = (delta >> 48) & 255
            output[j + 14] = (delta >> 56) & 255
            j += 15
        elif absdelta >= 1 << 15:
            output[j] = -128
            output[j + 1] = 0
            output[j + 2] = -128
            output[j + 3] = (delta & 255)
            output[j + 4] = (delta >> 8) & 255
            output[j + 5] = (delta >> 16) & 255
            output[j + 6] = (delta >> 24)
            j += 7
        elif absdelta >= 1 << 7:
            output[j] = -128
            output[j + 1] = delta & 255
            output[j + 2] = (delta >> 8) & 255
            j += 3
        else:
            output[j] = delta
            j += 1
        last = current
    return numpy.asarray(output)[:j]


@cython.boundscheck(False)
@cython.wraparound(False)
def dec_cbf(bytes stream not None, size=None):
    """
    Analyze a stream of char with any length of exception (2,4, or 8 bytes integers)
    :param stream: bytes (string) representing the compressed data
    :param size: the size of the output array (of longInts)
    :return: int64 ndArrays
    """
    cdef:
        int               i = 0
        int               j = 0
        uint8_t     tmp8 = 0

        int64_t    last = 0
        int64_t    current = 0
        int64_t    tmp64 = 0
        int64_t    tmp64a = 0
        int64_t    tmp64b = 0
        int64_t    tmp64c = 0
        int64_t    tmp64d = 0
        int64_t    tmp64e = 0
        int64_t    tmp64f = 0
        int64_t    tmp64g = 0

        uint8_t    key8 = 0x80
        uint8_t    key0 = 0x00

        int csize
        int lenStream = < int > len(stream)
        uint8_t[::1] cstream = bytearray(stream)
        int64_t[::1] data_out

    if size is None:
        csize = lenStream
    else:
        csize = < int > size

    data_out = numpy.empty(csize, dtype=numpy.int64)

    with nogil:
        while (i < lenStream) and (j < csize):
            if (cstream[i] == key8):
                if ((cstream[i + 1] == key0) and (cstream[i + 2] == key8)):
                    if (cstream[i + 3] == key0) and (cstream[i + 4] == key0) and (cstream[i + 5] == key0) and (cstream[i + 6] == key8):
                        # Retrieve the interesting Bytes of data
                        tmp64g = cstream[i + 7]
                        tmp64f = cstream[i + 8]
                        tmp64e = cstream[i + 9]
                        tmp64d = cstream[i + 10]
                        tmp64c = cstream[i + 11]
                        tmp64b = cstream[i + 12]
                        tmp64a = cstream[i + 13]
                        tmp64  = <int8_t> cstream[i + 14]
                        # Assemble data into a 64 bits integer
                        current = (tmp64 << 56) | (tmp64a << 48) | (tmp64b << 40) | (tmp64c << 32) | (tmp64d << 24) | (tmp64e << 16) | (tmp64f << 8) | (tmp64g)
                        i += 15
                    else:
                        # Retrieve the interesting Bytes of data
                        tmp64c = cstream[i + 3]
                        tmp64b = cstream[i + 4]
                        tmp64a = cstream[i + 5]
                        tmp64  = <int8_t> cstream[i + 6]
                        # Assemble data into a 64 bits integer
                        current = (tmp64 << 24) | (tmp64a << 16) | (tmp64b << 8) | (tmp64c)
                        i += 7
                else:
                    tmp64a = cstream[i + 1]
                    tmp64 = <int8_t> cstream[i + 2]

                    current = (tmp64 << 8) | (tmp64a)
                    i += 3
            else:
                current = (<int8_t> cstream[i])
                i += 1
            last += current
            data_out[j] = last
            j += 1

    return data_out[:j]


@cython.boundscheck(False)
def dec_cbf32(bytes stream not None, size=None):
    """
    Analyze a stream of char with any length of exception (2 or 4 bytes integers)
    Optimized for int32 decompression

    :param stream: bytes (string) representing the compressed data
    :param size: the size of the output array (of longInts)
    :return: int64 ndArrays
    """
    cdef:
        int               i = 0
        int               j = 0
        uint8_t     tmp8 = 0

        int32_t    last = 0
        int32_t    current = 0
        int32_t    tmp64 = 0
        int32_t    tmp64a = 0
        int32_t    tmp64b = 0
        int32_t    tmp64c = 0

        uint8_t    key8 = 0x80
        uint8_t    key0 = 0x00

        int csize
        int lenStream = < int > len(stream)
        uint8_t[:] cstream = bytearray(stream)
        int32_t[::1] data_out
    if size is None:
        csize = lenStream
    else:
        csize = < int > size
    data_out = numpy.empty(csize, dtype=numpy.int32)
    with nogil:
        while (i < lenStream) and (j < csize):
            if (cstream[i] == key8):
                if ((cstream[i + 1] == key0) and (cstream[i + 2] == key8)):
                    # Retrieve the interesting Bytes of data
                    tmp64c = cstream[i + 3]
                    tmp64b = cstream[i + 4]
                    tmp64a = cstream[i + 5]
                    tmp64  = <int8_t> cstream[i + 6]
                    # Assemble data into a 32 bits integer
                    current = (tmp64 << 24) | (tmp64a << 16) | (tmp64b << 8) | (tmp64c)
                    i += 7
                else:
                    tmp64a = cstream[i + 1]
                    tmp64  = <int8_t> cstream[i + 2]

                    current = (tmp64 << 8) | (tmp64a)
                    i += 3
            else:
                current = (<int8_t> cstream[i])
                i += 1
            last += current
            data_out[j] = last
            j += 1

    return numpy.asarray(data_out[:j])


@cython.boundscheck(False)
def dec_TY5(bytes stream not None, size=None):
    """
    Analyze a stream of char with a TY5 compression scheme and exception (2 or 4 bytes integers)

    TODO: known broken, FIXME

    :param stream: bytes (string) representing the compressed data
    :param size: the size of the output array (of longInts)
    :return: int32 ndArrays
    """

    cdef:
        int               i = 0
        int               j = 0
        int32_t    last = 0
        int32_t    current = 0
        uint8_t    key8 = 0xfe  # 127+127
        int32_t    tmp32a = 0
        int32_t    tmp32b = 0

        int csize
        int lenStream = len(stream)
        uint8_t[::1] cstream = bytearray(stream)
        int32_t[::1] data_out

    if size is None:
        csize = lenStream
    else:
        csize = < int > size

    data_out = numpy.zeros(csize, dtype=numpy.int32)
    if True:
        while (i < lenStream) and (j < csize):
            if (cstream[i] == key8):
                    tmp32a = cstream[i + 1] - 127
                    tmp32b = <int16_t>( <int8_t> cstream[i + 2] << 8 )
                    current = (tmp32b) | (tmp32a)
                    i += 3
            else:
                current = <int32_t>(<uint8_t> cstream[i]) - 127
                i += 1
            last += current
            data_out[j] = last
            j += 1
    return numpy.asarray(data_out[:j])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
"""
Test cases for the fabioimage class

testsuite by Jerome Kieffer (Jerome.Kieffer@esrf.eu)
28/11/2014
"""

import unittest
import os
import numpy
import copy
import logging
from ..fabioimage import FabioImage
from .. import fabioutils
from ..utils import pilutils
from .utilstest import UtilsTest

logger = logging.getLogger(__name__)

try:
    import pathlib
except ImportError:
    try:
        import pathlib2 as pathlib
    except ImportError:
        pathlib = None


class Test50000(unittest.TestCase):
    """test with 50000 everywhere"""

    def setUp(self):
        """make the image"""
        dat = numpy.ones((1024, 1024), numpy.uint16)
        dat = (dat * 50000).astype(numpy.uint16)
        assert dat.dtype.char == numpy.ones((1), numpy.uint16).dtype.char
        hed = {"Title": "50000 everywhere"}
        self.obj = FabioImage(dat, hed)

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        self.obj = None

    def testgetmax(self):
        """check max"""
        self.assertEqual(self.obj.getmax(), 50000)

    def testgetmin(self):
        """check min"""
        self.assertEqual(self.obj.getmin(), 50000)

    def testgetmean(self):
        """check mean"""
        self.assertEqual(self.obj.getmean(), 50000)

    def testgetstddev(self):
        """check stddev"""
        self.assertEqual(self.obj.getstddev(), 0)

    def testcopy(self):
        "test the copy statement"
        c = copy.copy(self.obj)
        self.assertNotEqual(id(c), id(self.obj), "object differ")
        self.assertEqual(c.header, self.obj.header, "header are the same")
        self.assertEqual(abs(c.data - self.obj.data).max(), 0, "data are the same")
        self.assertEqual(c.filename, self.obj.filename, "filename is the same")


class TestSlices(unittest.TestCase):
    """check slicing"""

    def setUp(self):
        """make test data"""
        dat2 = numpy.zeros((1024, 1024), numpy.uint16)
        hed = {"Title": "zeros and 100"}
        self.cord = [256, 256, 790, 768]
        self.obj = FabioImage(dat2, hed)
        self.slic = slic = self.obj.make_slice(self.cord)
        # Note - d2 is modified *after* fabioimage is made
        dat2[slic] = dat2[slic] + 100
        assert self.obj.maxval is None
        assert self.obj.minval is None
        self.npix = (slic[0].stop - slic[0].start) * (slic[1].stop - slic[1].start)

    def testgetmax(self):
        """check max"""
        self.assertEqual(self.obj.getmax(), 100)

    def testgetmin(self):
        """check min"""
        self.assertEqual(self.obj.getmin(), 0)

    def testintegratearea(self):
        """check integrations"""
        self.obj.resetvals()
        area1 = self.obj.integrate_area(self.cord)
        self.obj.resetvals()
        area2 = self.obj.integrate_area(self.slic)
        self.assertEqual(area1, area2)
        self.assertEqual(area1, self.npix * 100)

    def testRebin(self):
        """Test the rebin method"""
        big = numpy.arange(64).reshape((8, 8))
        res = numpy.array([[13, 17], [45, 49]])
        fabimg = FabioImage(data=big, header={})
        fabimg.rebin(4, 4)
        self.assertEqual(
            abs(res - fabimg.data).max(), 0, "data are the same after rebin"
        )


class TestOpen(unittest.TestCase):
    """check opening compressed files"""

    testfile = os.path.join(UtilsTest.tempdir, "testfile")

    def setUp(self):
        """create test files"""
        if not os.path.isfile(self.testfile):
            with open(self.testfile, "wb") as f:
                f.write(b"{ hello }")
        if not os.path.isfile(self.testfile + ".gz"):
            with fabioutils.GzipFile(self.testfile + ".gz", "wb") as wf:
                wf.write(b"{ hello }")
        if not os.path.isfile(self.testfile + ".bz2"):
            with fabioutils.BZ2File(self.testfile + ".bz2", "wb") as wf:
                wf.write(b"{ hello }")
        self.obj = FabioImage()

    def testFlat(self):
        """no compression"""
        res = self.obj._open(self.testfile)
        self.assertEqual(res.read(), b"{ hello }")
        res.close()

    def testgz(self):
        """gzipped"""
        res = self.obj._open(self.testfile + ".gz")
        self.assertEqual(res.read(), b"{ hello }")
        res.close()

    def testbz2(self):
        """bzipped"""
        res = self.obj._open(self.testfile + ".bz2")
        self.assertEqual(res.read(), b"{ hello }")
        res.close()

    def test_badtype(self):
        self.assertRaises(TypeError, self.obj._open, None)

    def test_pathlib(self):
        if pathlib is None:
            self.skipTest("pathlib is not available")
        path = pathlib.PurePath(self.testfile + ".bz2")
        res = self.obj._open(path)
        self.assertIsNotNone(res)
        res.close()


class TestPilImage(unittest.TestCase):
    """check PIL creation"""

    def setUp(self):
        if pilutils.Image is None:
            self.skipTest("PIL is not available")

        """ list of working numeric types"""
        self.okformats = [
            numpy.uint8,
            numpy.int8,
            numpy.uint16,
            numpy.int16,
            numpy.uint32,
            numpy.int32,
            numpy.float32,
        ]

    def mkdata(self, shape, typ):
        """generate [01] testdata"""
        return (numpy.random.random(shape)).astype(typ)

    def testpil(self):
        for typ in self.okformats:
            for shape in [(10, 20), (431, 1325)]:
                testdata = self.mkdata(shape, typ)
                img = FabioImage(testdata, {"title": "Random data"})
                pim = img.toPIL16()
                for i in [0, 5, 6, shape[1] - 1]:
                    for j in [0, 5, 7, shape[0] - 1]:
                        errstr = str(typ) + " %d %d %f %f t=%s" % (
                            i,
                            j,
                            testdata[j, i],
                            pim.getpixel((i, j)),
                            typ,
                        )

                        er1 = img.data[j, i] - pim.getpixel((i, j))
                        er2 = img.data[j, i] + pim.getpixel((i, j))

                        # difference as % error in case of rounding
                        if er2 != 0.0:
                            err = er1 / er2
                        else:
                            err = er1

                        self.assertAlmostEqual(err, 0, 6, errstr)


class TestPilImage2(TestPilImage):
    """check with different numbers"""

    def mkdata(self, shape, typ):
        """positive and big"""
        if numpy.issubdtype(typ, numpy.integer):
            maxi = numpy.iinfo(typ).max
        else:
            maxi = numpy.iinfo(int).max
        return (numpy.random.random(shape) * (0.1 * maxi)).astype(typ)


class TestPilImage3(TestPilImage):
    """check with different numbers"""

    def mkdata(self, shape, typ):
        """positive, negative and big"""
        if numpy.issubdtype(typ, numpy.integer):
            maxi = numpy.iinfo(typ).max
        else:
            maxi = numpy.iinfo(int).max
        return ((numpy.random.random(shape) - 0.5) * (0.1 * maxi)).astype(typ)


class TestDeprecatedFabioImage(unittest.TestCase):
    def test_patch_dim(self):
        data = numpy.array(numpy.arange(3 * 10)).reshape(3, 10)
        image = FabioImage(data=data)
        # Usecase found in some projects
        image.dim2, image.dim1 = data.shape
        # It should not change anything
        self.assertEqual(image.shape, data.shape)

    def test_cleanup_pilimage_cache(self):
        data = numpy.array(numpy.arange(3 * 10)).reshape(3, 10)
        image = FabioImage(data=data)
        # It was a way to force clean up of the cache
        image.pilimage = None


class TestFabioImage(unittest.TestCase):
    def test_iter_abort_iteration(self):
        data = numpy.zeros((2, 2))
        image = FabioImage(data=data)
        for frame in image:
            self.assertEqual(frame.data[0, 0], 0)

class TestSexedDtype(unittest.TestCase):
    def test_sexed_dtypes(self):
        a = FabioImage.get_stype("float32", ">")
        self.assertEqual(a.str, ">f4")
        a = FabioImage.get_stype("float64", "<")
        self.assertEqual(a.str, "<f8")
        a = FabioImage.get_stype(numpy.float32, "big")
        self.assertEqual(a.str, ">f4")
        a = FabioImage.get_stype(numpy.float64, "little")
        self.assertEqual(a.str, "<f8")

def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestFabioImage))
    testsuite.addTest(loadTests(Test50000))
    testsuite.addTest(loadTests(TestSlices))
    testsuite.addTest(loadTests(TestOpen))
    testsuite.addTest(loadTests(TestPilImage))
    testsuite.addTest(loadTests(TestPilImage2))
    testsuite.addTest(loadTests(TestPilImage3))
    testsuite.addTest(loadTests(TestDeprecatedFabioImage))
    testsuite.addTest(loadTests(TestSexedDtype))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
"""
# Unit tests

# builds on stuff from ImageD11.test.testpeaksearch
28/11/2014
"""

import unittest
import os
import logging
import fabio
import shutil
from .utilstest import UtilsTest

logger = logging.getLogger(__name__)


class TestHeaderNotSingleton(unittest.TestCase):
    def setUp(self):
        """
        download images
        """
        self.file1 = UtilsTest.getimage("mb_LP_1_001.img.bz2")[:-4]

    def testheader(self):
        file2 = self.file1.replace("mb_LP_1_001.img", "mb_LP_1_002.img")
        self.assertTrue(os.path.exists(self.file1))
        if not os.path.exists(file2):
            shutil.copy(self.file1, file2)
        image1 = fabio.open(self.file1)
        image2 = fabio.open(file2)

        self.assertEqual(self.abs_norm(image1.filename), self.abs_norm(self.file1))
        self.assertEqual(self.abs_norm(image2.filename), self.abs_norm(file2))
        self.assertNotEqual(image1.filename, image2.filename)

    def abs_norm(self, fn):
        return os.path.normcase(os.path.abspath(fn))

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        self.file1 = None


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestHeaderNotSingleton))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

"""FabIO module"""

__author__ = "Jérôme Kieffer"
__contact__ = "Jerome.Kieffer@ESRF.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "13/03/2026"
__status__ = "stable"

import sys
import logging
from .version import __date__ as date, version, version_info, hexversion, strictversion  # noqa
from . import fabioformats as _fabioformats
from . import fabioimage  # noqa
from . import openimage  # noqa
from .fabioutils import (
    jump_filename,  # noqa
    FilenameObject,  # noqa
    previous_filename,  # noqa
    next_filename,  # noqa
    deconstruct_filename,  # noqa
    extract_filenumber,  # noqa
    getnum,  # noqa
    construct_filename,  # noqa
    exists,  # noqa
)  # noqa
from .compression import COMPRESSORS  # noqa
from .openimage import openimage as open  # noqa
from .openimage import open_series as open_series  # noqa
from .openimage import openheader as openheader  # noqa

if "ps1" in dir(sys):
    # configure logging with interactive console
    logging.basicConfig()

# provide a global fabio API
factory = _fabioformats.factory

# feed the library with all the available formats
_fabioformats.register_default_formats()

# Compatibility with outside world:
filename_object = FilenameObject


def register(codec_class):
    """
    Register a codec class with the set of formats supported by fabio.

    It is a transitional function to prepare the next coming version of fabio.

    - On the current fabio library, when a module is imported, all the formats
        inheriting FabioImage are automatically registered. And this function is
        doing nothing.
    - On the next fabio library. Importing a module containing classes
        inheriting FabioImage will not be registered. And this function will
        register the class.

    The following source code will then provide the same behaviour on both
    fabio versions, and it is recommended to use it.

    .. code-block:: python

        @fabio.register
        class MyCodec(fabio.fabioimage.FabioImage):
            pass
    """
    assert issubclass(codec_class, fabioimage.FabioImage)
    _fabioformats.register(codec_class)
    return codec_class


def tests():
    """
    Run the FabIO test suite.

    If the test-images are not already installed (via the debian package for example),
    they need to be downloaded from sourceforge.net, which make take a while.
    Ensure your network connection is operational and your proxy settings are correct,
    for example:

    export http_proxy=http://proxy.site.com:3128
    """
    from . import test

    test.run_tests()


def benchmarks():
    """
    Run the benchmarks
    """
    from . import benchmark

    res = benchmark.run()
    return res



#!/usr/bin/env python3
import os
import sys
import logging
import bootstrap
from sphinx.cmd.build import main
logger = logging.getLogger(__name__)
PROJECT_DIR = os.path.dirname(os.path.abspath(__file__))
PROJECT_NAME = bootstrap.get_project_name(PROJECT_DIR)
logger.info("Project name: %s", PROJECT_NAME)
LIBPATH = bootstrap.build_project(PROJECT_NAME, PROJECT_DIR)
if __name__ == '__main__':
    sys.path.insert(0, LIBPATH)
    dest_dir = os.path.join(PROJECT_DIR, "build", "sphinx")
    if not os.path.isdir(dest_dir):
        os.makedirs(dest_dir)
    argv = ["-b", "html",
            os.path.join(PROJECT_DIR,"doc","source"),
            dest_dir ]
    print(f"run sphinx with those arguments {argv}")
    sys.exit(main(argv)) 



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""
# Unit tests for raxis images
28/11/2014
"""

__date__ = "27/10/2025"
import unittest
import os
import logging
import fabio
from fabio.raxisimage import raxisimage
from ..utilstest import UtilsTest

logger = logging.getLogger(__name__)

# filename dim1 dim2 min max mean stddev
TESTIMAGES = """mgzn-20hpt.img     2300 1280 16 15040  287.82 570.72
                mgzn-20hpt.img.bz2 2300 1280 16 15040  287.82 570.72
                mgzn-20hpt.img.gz  2300 1280 16 15040  287.82 570.72"""
#                Seek from end is not supported with gzip


class TestRaxisImage(unittest.TestCase):
    def setUp(self):
        """
        download images
        """
        self.mar = UtilsTest.getimage("mgzn-20hpt.img.bz2")[:-4]

    def test_read(self):
        """
        Test the reading of Mar345 images
        """
        for line in TESTIMAGES.split("\n"):
            vals = line.strip().split()
            name = vals[0]
            logger.debug("Testing file %s" % name)
            dim1, dim2 = [int(x) for x in vals[1:3]]
            shape = dim2, dim1
            mini, maxi, mean, stddev = [float(x) for x in vals[3:]]
            obj = raxisimage()
            obj.read(os.path.join(os.path.dirname(self.mar), name))

            self.assertAlmostEqual(
                mini, obj.getmin(), 2, "getmin [%s,%s]" % (mini, obj.getmin())
            )
            self.assertAlmostEqual(
                maxi, obj.getmax(), 2, "getmax [%s,%s]" % (maxi, obj.getmax())
            )
            self.assertAlmostEqual(
                mean, obj.getmean(), 2, "getmean [%s,%s]" % (mean, obj.getmean())
            )
            self.assertAlmostEqual(
                stddev,
                obj.getstddev(),
                2,
                "getstddev [%s,%s]" % (stddev, obj.getstddev()),
            )
            self.assertEqual(shape, obj.shape)

    def _test_write(self):
        self.skipTest("Write is not implemented")
        "Test writing with self consistency at the fabio level"
        for line in TESTIMAGES.split("\n"):
            logger.debug("Processing file: %s" % line)
            vals = line.split()
            name = vals[0]
            obj = raxisimage()
            obj.read(os.path.join(os.path.dirname(self.mar), name))
            obj.write(os.path.join(UtilsTest.tempdir, name))
            other = raxisimage()
            other.read(os.path.join(UtilsTest.tempdir, name))
            self.assertEqual(abs(obj.data - other.data).max(), 0, "data are the same")
            for key in obj.header:
                if key == "filename":
                    continue
                self.assertTrue(key in other.header, "Key %s is in header" % key)
                self.assertEqual(
                    obj.header[key],
                    other.header[key],
                    "value are the same for key %s: [%s|%s]"
                    % (key, obj.header[key], other.header[key]),
                )
            os.unlink(os.path.join(UtilsTest.tempdir, name))

    def test_memoryleak(self):
        """
        This test takes a lot of time, so only in debug mode.
        """
        N = 1000
        if logger.getEffectiveLevel() == logging.DEBUG:
            logger.debug("Testing for memory leak")
            for _ in range(N):
                _img = fabio.open(self.mar)


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestRaxisImage))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2026 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Helper to access to external resources.

Module imported from silx project to avoid cyclic dependency.
"""

__authors__ = ["Thomas Vincent", "J. Kieffer"]
__license__ = "MIT"
__date__ = "16/06/2026"


import hashlib
import json
import logging
import os
import sys
import tarfile
import tempfile
import threading
import time
import unittest
import urllib.request
import urllib.error
import zipfile
import filelock

try:
    import bz2
except ImportError:
    raise RuntimeError("BZ2 library is needed to decompress data")
    # Theoretically it would be possible to fall back on different compression scheme
try:
    import gzip
except ImportError:
    gzip = None


logger = logging.getLogger(__name__)


class ExternalResources:
    """Utility class which allows to download test-data from www.silx.org
    and manage the temporary data during the tests.

    """

    def __init__(self, project, url_base, env_key=None, timeout=60, data_home=None):
        """Constructor of the class

        :param str project: name of the project, like "silx"
        :param str url_base: base URL for the data, like "http://www.silx.org/pub"
        :param str env_key: name of the environment variable which contains the
                            test_data directory, like "SILX_DATA".
                            If None (default), then the name of the
                            environment variable is built from the project argument:
                            "<PROJECT>_DATA".
                            The environment variable is optional: in case it is not set,
                            a directory in the temporary folder is used.
        :param timeout: time in seconds before it breaks
        :param data_home: Directory in which the data will be downloaded
        """
        self.project = project
        self._initialized = False
        self.sem = threading.Semaphore()
        self.hash = hashlib.sha256
        self.env_key = env_key or (self.project.upper() + "_TESTDATA")
        self.url_base = url_base
        self.all_data = {}
        self.timeout = timeout
        self._data_home = data_home
        self.testdata = ""
        self.lock = None

    @property
    def _lockfile(self):
        """Returns the lockfile path."""
        return self.testdata + ".lock"

    @property
    def data_home(self):
        """Returns the data_home path and make sure it exists in the file
        system."""
        if self._data_home is not None:
            return self._data_home

        data_home = os.environ.get(self.env_key)
        if data_home is None:
            try:
                import getpass

                name = getpass.getuser()
            except Exception:
                if "getlogin" in dir(os):
                    name = os.getlogin()
                elif "USER" in os.environ:
                    name = os.environ["USER"]
                elif "USERNAME" in os.environ:
                    name = os.environ["USERNAME"]
                else:
                    name = "uid" + str(os.getuid())

            basename = f"{self.project}_testdata_{name}"
            data_home = os.path.join(tempfile.gettempdir(), basename)
        if not os.path.exists(data_home):
            try:
                os.makedirs(data_home, exist_ok=True)
            except OSError as err:
                raise RuntimeError(
                    f"Unable to create data directory {data_home} : {err}"
                )
        self._data_home = data_home

        return data_home

    def get_hash(self, filename=None, data=None):
        "Calculate and return the hash of a file or a bunch of data"
        if data is None and filename is None:
            return
        h = self.hash()
        if filename is not None:
            fullfilename = os.path.join(self.data_home, filename)
            if os.path.exists(fullfilename):
                with open(fullfilename, "rb") as fd:
                    data = fd.read()
            else:
                raise RuntimeError(f"Filename {fullfilename} does not exist !")
        h.update(data)
        return h.hexdigest()

    def _initialize_data(self):
        """Initialize for downloading test data"""
        if not self._initialized:
            with self.sem:
                if not self._initialized:
                    self.testdata = os.path.join(self.data_home, "all_testdata.json")
                    self.lock = filelock.FileLock(self._lockfile, timeout=self.timeout)
                    self.all_data = self.load_json()
                    self._initialized = True

    def clean_up(self):
        pass

    def getfile(self, filename):
        """Downloads the requested file from web-server available
        at https://www.silx.org/pub/silx/

        :param: relative name of the file.
        :return: full path of the locally saved file.
        """
        logger.debug(f"ExternalResources.getfile('{filename}')")

        if not self._initialized:
            self._initialize_data()

        fullfilename = os.path.abspath(os.path.join(self.data_home, filename))

        if os.path.isfile(fullfilename):
            if filename not in self.all_data:
                """File already exists but is not in the list of known files"""
                time_out = time.perf_counter() + self.timeout
                while time.perf_counter() < time_out:
                    dico = self.load_json()
                    if filename in dico:
                        dico.update(self.all_data)
                        self.all_data = dico
                        break
                    time.sleep(1)
                else:
                    logger.warning(
                        f"Timeout! Filename {filename} not present in all_data:{os.linesep}{json.dumps(self.all_data, indent=2)}"
                    )
                    os.remove(fullfilename)
                    return self.getfile(filename)

            h = self.hash()
            try:
                with self.lock:
                    with open(fullfilename, mode="rb") as fd:
                        h.update(fd.read())
            except filelock.Timeout:
                logger.error(f"Unable to lock to read {filename} file")

            if h.hexdigest() != self.all_data[filename]:
                logger.warning(f"Detected corrupted file {fullfilename} !")
                self.all_data.pop(filename)
                os.unlink(fullfilename)
                return self.getfile(filename)
        else:
            logger.debug(
                f"Trying to download file {filename}, timeout set to {self.timeout}s",
            )
            dictProxies = {}
            if "http_proxy" in os.environ:
                dictProxies["http"] = os.environ["http_proxy"]
                dictProxies["https"] = os.environ["http_proxy"]
            if "https_proxy" in os.environ:
                dictProxies["https"] = os.environ["https_proxy"]
            if dictProxies:
                proxy_handler = urllib.request.ProxyHandler(dictProxies)
                opener = urllib.request.build_opener(proxy_handler).open
            else:
                opener = urllib.request.urlopen

            logger.debug(f"wget {self.url_base}/{filename}")
            try:
                data = opener(
                    f"{self.url_base}/{filename}", data=None, timeout=self.timeout
                ).read()
                logger.info(f"File {filename} successfully downloaded.")
            except urllib.error.URLError:
                raise unittest.SkipTest("network unreachable.")

            dirname = os.path.dirname(fullfilename)

            if not os.path.exists(dirname):
                try:
                    os.makedirs(dirname, exist_ok=True)
                except OSError as err:
                    raise RuntimeError(
                        f"Unable to create directory {dirname}: {err}"
                    )

            try:
                with self.lock, open(fullfilename, mode="wb") as outfile:
                    outfile.write(data)
            except filelock.Timeout:
                logger.error(f"Unable to lock to write {filename} file.")
            except OSError:
                raise OSError(f"unable to write downloaded \
                    data into {filename} file.")

            if os.path.isfile(fullfilename):
                self.all_data[filename] = self.get_hash(data=data)
                self.save_json()
            else:
                raise RuntimeError(f"""Could not automatically download test files {filename}!
                    If you are behind a firewall, please set both environment variable
                     http_proxy and https_proxy.
                    This even works under windows !
                    Otherwise please try to download the files manually from
                    {self.url_base}/{filename}""")
        return fullfilename

    get_file = getfile

    def load_json(self) -> dict:
        """Loads the JSON file containing the list of files and their hashes

        :return: dict with filename:hash
        """
        all_data = {}
        if self.testdata and os.path.exists(self.testdata):
            try:
                with self.lock, open(self.testdata) as fd:
                        jdata = json.load(fd)
            except filelock.Timeout:
                logger.error("Unable to lock JSON file")
                jdata = {}
            if isinstance(jdata, dict):
                all_data = jdata
            else:
                # recalculate the hash only if the data was stored as a list
                all_data = {k: self.get_hash(k) for k in jdata}
        return all_data

    def save_json(self):
        """Saves the JSON file containing the list of files and their hashes"""
        dico = self.load_json()
        dico.update(self.all_data)
        file_list = list(dico.keys())
        file_list.sort()
        dico = {i: dico[i] for i in file_list}  # reorder items

        try:
            with self.lock:
                with open(self.testdata, "w") as fp:
                    json.dump(dico, fp, indent=4)
        except filelock.Timeout:
            logger.error("Unable to lock JSON file")
        except OSError:
            logger.error("Unable to save JSON dict")

    def getdir(self, dirname:str) -> list:
        """Downloads the requested tarball from the server
        https://www.silx.org/pub/silx/
        and unzips it into the data directory

        :param: relative name of the file.
        :return: list of files with their full path.
        """
        lodn = dirname.lower()
        full_path = self.getfile(dirname)
        output = os.path.join(self.data_home, dirname + "__content")

        if lodn.endswith(("tar", "tgz", "tbz2", "tar.gz", "tar.bz2")):
            with tarfile.TarFile.open(full_path, mode="r") as fd:
                # Avoid unsafe filter deprecation warning during mode change
                if (3, 12) <= sys.version_info < (3, 14):
                    fd.extraction_filter = tarfile.data_filter
                fd.extractall(output)
                return [os.path.join(output, i) for i in fd.getnames()]

        if lodn.endswith("zip"):
            with zipfile.ZipFile(full_path, mode="r") as fd:
                fd.extractall(output)
                return [os.path.join(output, i) for i in fd.namelist()]

        raise RuntimeError(
            "Unsupported archive format. Only tar and zip are currently supported"
        )

    get_dir = getdir

    def get_file_and_repack(self, filename:str) -> str:
        """
        Download the bzipped2 file, decompress it and repack it to gzip, if needed

        :param str filename: name of the file.
        :rtype: str
        :return: full path of the locally saved file
        """
        if not self._initialized:
            self._initialize_data()


        basefilename = os.path.basename(filename)
        fullfilename = os.path.abspath(os.path.join(self.data_home, basefilename))

        if filename in self.all_data and os.path.isfile(fullfilename):
            if self.all_data[filename] == self.get_hash(filename):
                return fullfilename

        # Else we download them all and decompress/recompress.
        if basefilename.endswith(".bz2"):
            bzip2name = basefilename
            basename = basefilename[:-4]
            gzipname = basename + ".gz"
        elif basefilename.endswith(".gz"):
            gzipname = basefilename
            basename = basefilename[:-3]
            bzip2name = basename + ".bz2"
        else:
            basename = basefilename
            gzipname = basefilename + ".gz"
            bzip2name = basename + ".bz2"

        fullfilename_gz = os.path.abspath(os.path.join(self.data_home, gzipname))
        fullfilename_raw = os.path.abspath(os.path.join(self.data_home, basename))
        fullfilename_bz2 = os.path.abspath(os.path.join(self.data_home, bzip2name))

        # The files are recreated from the bz2 file
        if not os.path.isfile(fullfilename_bz2):
            self.getfile(bzip2name)
            if not os.path.isfile(fullfilename_bz2):
                raise RuntimeError(f"""Could not automatically download test file: {filename}!
                    If you are behind a firewall, please set the environment variable
                    http_proxy.
                    Otherwise please try to download the files manually from
                    {self.url_base}""")

        raw_file_exists = os.path.isfile(fullfilename_raw)
        gz_file_exists = os.path.isfile(fullfilename_gz)
        if not raw_file_exists or not gz_file_exists:
            try:
                with self.lock, open(fullfilename_bz2, "rb") as fd:
                        data = fd.read()
            except filelock.Timeout:
                logger.error(f"Unable to obtain lock to read {bzip2name} file.")

            decompressed = bz2.decompress(data)

            if not raw_file_exists:
                try:
                    with self.lock, open(fullfilename_raw, "wb") as fd:
                            fd.write(decompressed)
                except filelock.Timeout:
                    logger.error(f"Unable to obtain lock to write {basename} file.")
                except OSError:
                    raise OSError(f"Unable to write decompressed data to disk at {self.data_home}")
                else:
                    self.all_data[basename] = self.get_hash(basename)

            if not gz_file_exists:
                if gzip is None:
                    logger.error("Gzip library expected to recompress data")
                try:
                    with self.lock, gzip.open(fullfilename_gz, "wb") as fd:
                            fd.write(decompressed)
                except filelock.Timeout:
                    logger.error(f"Unable to obtain lock to write {gzipname} file.")
                except OSError as err:
                    raise OSError(f"Unable to write gzipped data to disk at {self.data_home}: {err}")
                else:
                    self.all_data[gzipname] = self.get_hash(gzipname)
        return fullfilename

    def download_all(self, imgs:list|None=None) -> list:
        """Download all data needed for the test/benchmarks

        :param imgs: list of files to download, by default all
        :return: list of path of all files
        """
        if not self._initialized:
            self._initialize_data()
        if not imgs:
            imgs = self.all_data.keys()
        res = []
        for fn in imgs:
            logger.info(f"Downloading from silx.org: {fn}")
            res.append(self.getfile(fn))
        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
"""
# Unit tests

builds on stuff from ImageD11.test.testpeaksearch
Jerome Kieffer 04/12/2014
"""

import unittest
import logging
import io
from fabio.openimage import openimage
from fabio.edfimage import EdfImage
from fabio.marccdimage import MarccdImage
from fabio.fit2dmaskimage import Fit2dMaskImage
from fabio.OXDimage import OXDimage
from fabio.brukerimage import BrukerImage
from fabio.dtrekimage import DtrekImage
from .utilstest import UtilsTest

logger = logging.getLogger(__name__)


class TestOpenEdf(unittest.TestCase):
    """openimage opening edf"""

    def checkFile(self, filename):
        """check we can read EDF image with openimage"""
        obj = openimage(filename)
        obj2 = EdfImage()
        obj2.read(filename)
        self.assertEqual(obj.data[10, 10], obj2.data[10, 10])
        self.assertEqual(type(obj), type(obj2))
        self.assertEqual(abs(obj.data.astype(int) - obj2.data.astype(int)).sum(), 0)

    def testEdf(self):
        fname = "F2K_Seb_Lyso0675.edf.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        self.checkFile(filename)

    def testEdfGz(self):
        fname = "F2K_Seb_Lyso0675.edf.gz"
        filename = UtilsTest.getimage(fname)
        self.checkFile(filename)

    def testEdfBz2(self):
        fname = "F2K_Seb_Lyso0675.edf.bz2"
        filename = UtilsTest.getimage(fname)
        self.checkFile(filename)

    def testStream(self):
        fname = "F2K_Seb_Lyso0675.edf.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        with open(filename, "rb") as fd:
            stream = io.BytesIO(fd.read())
        fimg = openimage(stream)
        ref = EdfImage()
        ref.read(filename)

        self.assertEqual(type(ref), type(fimg), "type matches")
        self.assertEqual(ref.shape, fimg.shape, "shape matches")
        self.assertEqual(ref.dtype, fimg.dtype, "dtype matches")


class TestOpenMccd(unittest.TestCase):
    """openimage opening mccd"""

    def checkFile(self, filename):
        """check we can read it"""
        obj = openimage(filename)
        obj2 = MarccdImage()
        obj2.read(filename)
        self.assertEqual(obj.data[10, 10], obj2.data[10, 10])
        self.assertEqual(type(obj), type(obj2))
        self.assertEqual(abs(obj.data.astype(int) - obj2.data.astype(int)).sum(), 0)

    def testMccd(self):
        fname = "somedata_0001.mccd.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        self.checkFile(filename)

    def testMccdGz(self):
        fname = "somedata_0001.mccd.gz"
        filename = UtilsTest.getimage(fname)
        self.checkFile(filename)

    def testMccdBz2(self):
        fname = "somedata_0001.mccd.bz2"
        filename = UtilsTest.getimage(fname)
        self.checkFile(filename)

    def testStream(self):
        fname = "somedata_0001.mccd.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        with open(filename, "rb") as fd:
            stream = io.BytesIO(fd.read())
        fimg = openimage(stream)
        ref = MarccdImage()
        ref.read(filename)

        self.assertTrue("TifImage" in type(fimg).__name__, "type matches")
        self.assertEqual(ref.shape, fimg.shape, "shape matches")
        self.assertEqual(ref.dtype, fimg.dtype, "dtype matches")


class TestOpenMask(unittest.TestCase):
    """openimage opening fit2d msk"""

    def checkFile(self, filename):
        """check we can read Fit2D mask with openimage"""
        obj = openimage(filename)
        obj2 = Fit2dMaskImage()
        obj2.read(filename)
        self.assertEqual(obj.data[10, 10], obj2.data[10, 10])
        self.assertEqual(type(obj), type(obj2))
        self.assertEqual(abs(obj.data.astype(int) - obj2.data.astype(int)).sum(), 0)
        self.assertEqual(abs(obj.data.astype(int) - obj2.data.astype(int)).sum(), 0)

    def testMask(self):
        """openimage opening fit2d msk"""
        fname = "face.msk.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        self.checkFile(filename)

    def testMaskGz(self):
        """openimage opening fit2d msk gzip"""
        fname = "face.msk.gz"
        filename = UtilsTest.getimage(fname)
        self.checkFile(filename)

    def testMaskBz2(self):
        """openimage opening fit2d msk bzip"""
        fname = "face.msk.bz2"
        filename = UtilsTest.getimage(fname)
        self.checkFile(filename)

    def testStream(self):
        fname = "face.msk.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        with open(filename, "rb") as fd:
            stream = io.BytesIO(fd.read())
        fimg = openimage(stream)
        ref = Fit2dMaskImage()
        ref.read(filename)

        self.assertTrue("Fit2dMaskImage" in type(fimg).__name__, "type matches")
        self.assertEqual(ref.shape, fimg.shape, "shape matches")
        self.assertEqual(ref.dtype, fimg.dtype, "dtype matches")

class TestOpenBruker(unittest.TestCase):
    """openimage opening bruker"""

    def checkFile(self, filename):
        """check we can read it"""
        obj = openimage(filename)
        obj2 = BrukerImage()
        obj2.read(filename)
        self.assertEqual(obj.data[10, 10], obj2.data[10, 10])
        self.assertEqual(type(obj), type(obj2))
        self.assertEqual(abs(obj.data.astype(int) - obj2.data.astype(int)).sum(), 0)

    def testBruker(self):
        """openimage opening bruker"""
        fname = "Cr8F8140k103.0026.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        self.checkFile(filename)

    def testBrukerGz(self):
        """openimage opening bruker gzip"""
        fname = "Cr8F8140k103.0026.gz"
        filename = UtilsTest.getimage(fname)
        self.checkFile(filename)

    def testBrukerBz2(self):
        """openimage opening bruker bzip"""
        fname = "Cr8F8140k103.0026.bz2"
        filename = UtilsTest.getimage(fname)
        self.checkFile(filename)

    def testStream(self):
        fname = "Cr8F8140k103.0026.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        with open(filename, "rb") as fd:
            stream = io.BytesIO(fd.read())
        fimg = openimage(stream)
        ref = BrukerImage()
        ref.read(filename)

        self.assertTrue("BrukerImage" in type(fimg).__name__, "type matches")
        self.assertEqual(ref.shape, fimg.shape, "shape matches")
        self.assertEqual(ref.dtype, fimg.dtype, "dtype matches")


class TestOpenDtrek(unittest.TestCase):
    """openimage opening adsc"""

    def checkFile(self, filename):
        """check we can read it"""
        obj = openimage(filename)
        obj2 = DtrekImage()
        obj2.read(filename)
        self.assertEqual(obj.data[10, 10], obj2.data[10, 10])
        self.assertEqual(type(obj), type(obj2))
        self.assertEqual(abs(obj.data.astype(int) - obj2.data.astype(int)).sum(), 0)

    def testAdsc(self):
        """openimage opening adsc"""
        fname = "mb_LP_1_001.img.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        self.checkFile(filename)

    def testAdscGz(self):
        """openimage opening adsc gzip"""
        fname = "mb_LP_1_001.img.gz"
        filename = UtilsTest.getimage(fname)
        self.checkFile(filename)

    def testAdscBz2(self):
        """openimage opening adsc bzip"""
        fname = "mb_LP_1_001.img.bz2"
        filename = UtilsTest.getimage(fname)
        self.checkFile(filename)

    def testStream(self):
        fname = "mb_LP_1_001.img.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        with open(filename, "rb") as fd:
            stream = io.BytesIO(fd.read())
        fimg = openimage(stream)
        ref = DtrekImage()
        ref.read(filename)

        self.assertTrue("DtrekImage" in type(fimg).__name__, "type matches")
        self.assertEqual(ref.shape, fimg.shape, "shape matches")
        self.assertEqual(ref.dtype, fimg.dtype, "dtype matches")


class TestOpenOxd(unittest.TestCase):
    """openimage opening adsc"""

    def checkFile(self, filename):
        """check we can read OXD images with openimage"""
        obj = openimage(filename)
        obj2 = OXDimage()
        obj2.read(filename)
        self.assertEqual(obj.data[10, 10], obj2.data[10, 10])
        self.assertEqual(type(obj), type(obj2))
        self.assertEqual(abs(obj.data.astype(int) - obj2.data.astype(int)).sum(), 0)

    def testOxd(self):
        """openimage opening adsc"""
        fname = "b191_1_9_1.img.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        self.checkFile(filename)

    def testOxdUnc(self):
        """openimage opening adsc"""
        fname = "b191_1_9_1_uncompressed.img.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        self.checkFile(filename)

    def testStream(self):
        fname = "b191_1_9_1_uncompressed.img.bz2"
        filename = UtilsTest.getimage(fname)[:-4]
        with open(filename, "rb") as fd:
            stream = io.BytesIO(fd.read())
        fimg = openimage(stream)
        ref = OXDimage()
        ref.read(filename)

        self.assertTrue("OxdImage" in type(fimg).__name__, "type matches")
        self.assertEqual(ref.shape, fimg.shape, "shape matches")
        self.assertEqual(ref.dtype, fimg.dtype, "dtype matches")


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestOpenDtrek))
    testsuite.addTest(loadTests(TestOpenBruker))
    testsuite.addTest(loadTests(TestOpenEdf))
    testsuite.addTest(loadTests(TestOpenMask))
    testsuite.addTest(loadTests(TestOpenMccd))
    testsuite.addTest(loadTests(TestOpenOxd))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
"""

Authors: Henning O. Sorensen & Erik Knudsen
         Center for Fundamental Research: Metal Structures in Four Dimensions
         Risoe National Laboratory
         Frederiksborgvej 399
         DK-4000 Roskilde
         email:erik.knudsen@risoe.dk

+ mods for fabio by JPW

"""

import numpy
import re

import logging

from .fabioimage import FabioImage
from .fabioutils import ENDIANNESS

logger = logging.getLogger(__name__)

_DATA_TYPES = {
    "signed char": numpy.int8,
    "unsigned char": numpy.uint8,
    "short int": numpy.int16,
    "unsigned short int": numpy.uint16,
    "long int": numpy.int32,
    "unsigned long int": numpy.uint32,
    "float IEEE": numpy.float32,
    # Valid but unsupported
    "Compressed": None,
    # Valid but unsupported
    "Other_type": None,
}
"""Mapping from Data_type content to numpy equivalent"""


class DtrekImage(FabioImage):
    """Read an image using the d*TREK format.

    This format is used to process X-ray diffraction data from area detectors.
    It supports processing of data from multiple detector types (imaging plates,
    CCDs and pixel arrays) and from multiple vendors (Rigaku, Mar, Dectris,
    Bruker and ADSC).

    Rigaku providing a `specification <https://www.rigaku.com/downloads/software/free/dTREK%20Image%20Format%20v1.1.pdf>`_.
    """

    DESCRIPTION = "D*trek format (Rigaku specification 1.1)"

    DEFAULT_EXTENSIONS = ["img"]

    _keyvalue_pattern = None

    def __init__(self, *args, **kwargs):
        FabioImage.__init__(self, *args, **kwargs)
        if DtrekImage._keyvalue_pattern is None:
            DtrekImage._keyvalue_pattern = re.compile(b"[^\n]+")

    def read(self, fname, frame=None):
        """read in the file"""
        with self._open(fname, "rb") as infile:
            try:
                self._readheader(infile)
            except Exception:
                logger.debug("Backtrace", exc_info=True)
                raise IOError("Error processing d*TREK header")

            # FIXME: It would be good to read only the expected data
            binary = infile.read()

        # Read information of the binary data type
        data_type = self.header.get("Data_type", None)
        if data_type is None:
            # Compatibility with old supported files
            data_type = self.header.get("TYPE", None)
            if data_type is not None and data_type == "unsigned_short":
                pass
            else:
                logger.warning(
                    "Data_type key is mandatory. Fallback to unsigner integer 16-bits."
                )
            numpy_type = numpy.uint16
        else:
            if data_type not in _DATA_TYPES:
                raise IOError(
                    "Data_type key contains an invalid/unsupported value: %s", data_type
                )
            numpy_type = _DATA_TYPES[data_type]
            if type is None:
                raise IOError("Data_type %s is not supported by fabio", data_type)

        # Stored in case data reading fails
        self._dtype = numpy.dtype(numpy_type)

        dim = self.header.get("DIM", None)
        if dim is None:
            logger.warning("DIM key is mandatory. Fallback using DIM=2.")
            dim = 2
        else:
            dim = int(dim)

        shape = []
        for i in range(dim):
            value = int(self.header["SIZE%d" % (i + 1)])
            shape.insert(0, value)
        self._shape = shape

        if sum(shape) == 0:
            data = None
        else:
            # Read the data into the array
            stype = self.get_stype(self._dtype, self._get_dtrek_byte_order())
            data = numpy.frombuffer(binary, stype).astype(self._dtype)
            try:
                data = data.reshape(self._shape)
            except ValueError:
                raise IOError(
                    "Size spec in d*TREK header does not match "
                    + "size of image data field %s != %s" % (self._shape, data.size)
                )
        self.data = data
        self._shape = None
        self._dtype = None
        self.resetvals()
        return self

    def _split_meta(self, line):
        """Split a line into key and value.

        :param bytes line: A line of bytes
        :rtype: Tuple[str,str]
        """
        if b"=" not in line:
            raise ValueError("No meta")
        line = line.decode("ascii")
        key, value = line.split("=")
        return key.strip(), value.strip(" ;\n\r")

    def _readheader(self, infile):
        """Read a d*TREK header.

        After the execution of this function, the cursor on infile will point
        at the end of the header (at the start of the binary data block).

        :param FileObject infile: A file object pointing at the first character
            of the header.
        """
        header_line = infile.readline()
        assert header_line.startswith(b"{")
        header_bytes_line = infile.readline()
        key, header_bytes = self._split_meta(header_bytes_line)
        assert key == "HEADER_BYTES"
        self.header[key] = header_bytes
        header_bytes = int(header_bytes)

        # Read the remining block
        # For robustness, cause that's in fact a const
        header_bytes -= len(header_line) + len(header_bytes_line)
        header_block = infile.read(header_bytes)

        for line in DtrekImage._keyvalue_pattern.finditer(header_block):
            line = line.group(0)
            if line.startswith(b"}"):
                # Remining part is padding
                return
            try:
                key, value = self._split_meta(line)
                self.header[key] = value
            except ValueError:
                pass

        # It means there was no end of block
        logger.warning("The end of block '}' was not reachable. File may be corrupted.")

    def write(self, fname):
        """
        Write d*TREK format
        """

        # From specification
        HEADER_START = b"{\n"
        HEADER_END = b"}\n\x0c\n"
        HEADER_BYTES_TEMPLATE = "HEADER_BYTES=% 5d;\n"
        # start + end + header_bytes_key + header_bytes_value + header_bytes_end
        MINIMAL_HEADER_SIZE = 2 + 4 + 13 + 5 + 2

        data = self.data
        if data is not None:
            dtrek_data_type = None
            for key, value in _DATA_TYPES.items():
                if data.dtype.type == value:
                    dtrek_data_type = key
                    break

            if dtrek_data_type is None:
                if data.dtype.kind == "f":
                    dtrek_data_type = "float IEEE"
                elif data.dtype.kind == "u":
                    dtrek_data_type = "unsigned long int"
                elif data.dtype.kind == "i":
                    dtrek_data_type = "long int"
                else:
                    raise TypeError("Unsupported data type %s", data.dtype)
                new_dtype = numpy.dtype(_DATA_TYPES[dtrek_data_type])
                logger.warning(
                    "Data type %s unsupported. Store it as %s.", data.dtype, new_dtype
                )
                data = data.astype(new_dtype)

            byte_order = self._get_dtrek_byte_order(
                default=ENDIANNESS.LITTLE if numpy.little_endian else ENDIANNESS.BIG)
            data = data.astype(self.get_stype(data.dtype, byte_order))

            # Patch header to match the data
            self.header["Data_type"] = dtrek_data_type
            self.header["DIM"] = str(len(data.shape))
            for i, size in enumerate(reversed(data.shape)):
                self.header["SIZE%d" % (i + 1)] = str(size)
            self.header["BYTE_ORDER"] = "little_endian" if byte_order==ENDIANNESS.LITTLE else "big_endian"
        else:
            # No data
            self.header["Data_type"] = "long int"
            self.header["DIM"] = "2"
            self.header["SIZE1"] = "0"
            self.header["SIZE2"] = "0"
            self.header["BYTE_ORDER"] = "little_endian"

        out = b""
        for key in self.header:
            if key == "HEADER_BYTES":
                continue
            line = "%s= %s;\n" % (key, self.header[key])
            out += line.encode("utf-8")

        # FIXME: This code do not take into account the size of "HEADER_BYTES"
        if "HEADER_BYTES" in self.header:
            hsize = int(self.header["HEADER_BYTES"])
            pad = hsize - len(out) - MINIMAL_HEADER_SIZE
            if pad < 0:
                logger.warning("HEADER_BYTES have to be patched.")
                minimal_hsize = hsize - pad
                hsize = (minimal_hsize + 512) & ~(512 - 1)
                pad = hsize - minimal_hsize
        else:
            minimal_hsize = len(out) + MINIMAL_HEADER_SIZE
            hsize = (minimal_hsize + 512) & ~(512 - 1)
            pad = hsize - minimal_hsize

        header_bytes = HEADER_BYTES_TEMPLATE % hsize
        out = (
            HEADER_START
            + header_bytes.encode("ascii")
            + out
            + HEADER_END
            + (b" " * pad)
        )
        assert len(out) % 512 == 0, "Header is not multiple of 512"

        with open(fname, "wb") as outf:
            outf.write(out)
            if data is not None:
                data.tofile(outf)

    def _get_dtrek_byte_order(self, default=ENDIANNESS.LITTLE) -> ENDIANNESS:
        """Returns the byte order value in d*TREK format."""
        if "BYTE_ORDER" in self.header:
            byte_order = self.header["BYTE_ORDER"]
            if "little" in byte_order:
                return ENDIANNESS.LITTLE
            elif "big" in byte_order:
                return ENDIANNESS.BIG
            else:
                logger.warning(
                    "Invalid BYTE_ORDER value. Found '%s', assuming default: %s",
                    byte_order, default
                )
                return default
        else:
            return default



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
"""
# Unit tests
Jerome Kieffer, 04/12/2014
"""

import unittest
import logging
from fabio.openimage import openheader
from .utilstest import UtilsTest

logger = logging.getLogger(__name__)


class Test1(unittest.TestCase):
    """openheader opening edf"""

    def setUp(self):
        self.name = UtilsTest.getimage("F2K_Seb_Lyso0675_header_only.edf.bz2")[:-4]

    def testcase(self):
        """check openheader can read edf headers"""
        for ext in ["", ".bz2", ".gz"]:
            name = self.name + ext
            obj = openheader(name)
            logger.debug(" %s obj = %s" % (name, obj.header))
            self.assertEqual(
                obj.header["title"], "ESPIA FRELON Image", "Error on file %s" % name
            )


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(Test1))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite)



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

"""
Test module FabIO
"""

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "15/06/2026"

import unittest
from . import test_all


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(test_all.suite())
    return test_suite


def run_tests():
    """Run test complete test_suite"""
    mysuite = test_all.suite()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        print("Test suite failed")
        return 1
    else:
        print("Test suite succeeded")
        return 0



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
"""
Test cases for the flat binary images

testsuite by Jerome Kieffer (Jerome.Kieffer@esrf.eu)
28/11/2014
"""

import unittest
import os
import logging
from .utilstest import UtilsTest
import fabio

logger = logging.getLogger(__name__)


class TestFlatBinary(unittest.TestCase):
    filenames = [
        os.path.join(UtilsTest.tempdir, i)
        for i in (
            "not.a.file",
            "bad_news_1234",
            "empty_files_suck_1234.edf",
            "notRUBY_1234.dat",
        )
    ]

    def setUp(self):
        for filename in self.filenames:
            with open(filename, "wb") as f:
                # A 2048 by 2048 blank image
                f.write("\0x0" * 2048 * 2048 * 2)

    def test_openimage(self):
        """
        test the opening of "junk" empty images ...
        JK: I wonder if this test makes sense !
        """
        nfail = 0
        for filename in self.filenames:
            try:
                im = fabio.open(filename)
                if im.data.tobytes() != "\0x0" * 2048 * 2048 * 2:
                    nfail += 1
                else:
                    logger.info("**** Passed: %s" % filename)
            except Exception:
                logger.warning("failed for: %s" % filename)
                nfail += 1
        self.assertEqual(
            nfail, 0, " %s failures out of %s" % (nfail, len(self.filenames))
        )

    def tearDown(self):
        for filename in self.filenames:
            os.remove(filename)


def suite():
    # loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    # testsuite.addTest(loadTests(TestFlatBinary))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

"""
Test the main fabio API.

Basically everything supposed to be provided by `import fabio`
"""

import unittest
import logging
import io
from .utilstest import UtilsTest
import fabio

logger = logging.getLogger(__name__)


class TestFabio(unittest.TestCase):
    def test_open(self):
        filename = UtilsTest.getimage("multiframes.edf.bz2")
        filename = filename.replace(".bz2", "")
        image = fabio.open(filename)
        image.data
        image.close()

    def test_open_bytesio(self):
        filename = UtilsTest.getimage("multiframes.edf.bz2")
        filename = filename.replace(".bz2", "")
        with io.open(filename, "rb") as f:
            data = f.read()
            mem = io.BytesIO(data)
            with fabio.open(mem) as image:
                self.assertIsNotNone(image)
                self.assertEqual(image.nframes, 8)

    def test_open_fabio_bytesio(self):
        filename = UtilsTest.getimage("multiframes.edf.bz2")
        filename = filename.replace(".bz2", "")
        with io.open(filename, "rb") as f:
            data = f.read()
            mem = fabio.fabioutils.BytesIO(data)
            with fabio.open(mem) as image:
                self.assertIsNotNone(image)
                self.assertEqual(image.nframes, 8)

    def test_open_with(self):
        filename = UtilsTest.getimage("multiframes.edf.bz2")
        filename = filename.replace(".bz2", "")
        with fabio.open(filename) as image:
            image.data

    def test_open_series(self):
        filename = UtilsTest.getimage("multiframes.edf.bz2")
        filename = filename.replace(".bz2", "")
        series = fabio.open_series(filenames=[filename])
        for _frame in series.frames():
            pass
        series.close()

    def test_open_series_with(self):
        filename = UtilsTest.getimage("multiframes.edf.bz2")
        filename = filename.replace(".bz2", "")
        with fabio.open_series(filenames=[filename]) as series:
            for _frame in series.frames():
                pass

    def test_open_series_first_filename(self):
        filename = UtilsTest.getimage("multiframes.edf.bz2")
        filename = filename.replace(".bz2", "")
        with fabio.open_series(first_filename=filename) as series:
            for _frame in series.frames():
                pass


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestFabio))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python
import sys
import numpy
print("Python %s bits" % (tuple.__itemsize__ * 8))
print("       maxsize: %s\t maxunicode: %s" % (sys.maxsize, sys.maxunicode))
print(sys.version)
try:
    from distutils.sysconfig import get_config_vars
except:
    from sysconfig import get_config_vars
print("Config "+" ".join(get_config_vars("CONFIG_ARGS")))
print()
print("Numpy %s" % numpy.version.version)
print("      include %s" % numpy.get_include())
print("      options %s" % numpy.get_printoptions())



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
Reads the header from a GE a-Si Angio Detector
using version 8001 of the header from file:

    c:\\adept\\core\\DefaultImageInfoConfig.csv
"""

__authors__ = ["Antonino Miceli", "Jon Wright", "Jérôme Kieffer", "Joel Bernier"]
__date__ = "17/06/2026"
__status__ = "production"
__copyright__ = "2007-2020 APS; 2010-2020 ESRF"
__licence__ = "MIT"

import io
import numpy
import struct
from .edfimage import EdfImage
from .fabioimage import FabioImage
from .fabioutils import next_filename, previous_filename
from .openimage import MAGIC_NUMBERS

EDF_MAGIC_NUMBERS = [(x, y) for x, y in MAGIC_NUMBERS if y == "edf"]

GE_HEADER_INFO = [
    # Name, length in bytes, format for struct (None means string)
    ("ImageFormat", 10, None),
    ("VersionOfStandardHeader", 2, "<H"),
    ("StandardHeaderSizeInBytes", 4, "<L"),
    ("VersionOfUserHeader", 2, "<H"),
    ("UserHeaderSizeInBytes", 4, "<L"),
    ("NumberOfFrames", 2, "<H"),
    ("NumberOfRowsInFrame", 2, "<H"),
    ("NumberOfColsInFrame", 2, "<H"),
    ("ImageDepthInBits", 2, "<H"),
    ("AcquisitionDate", 20, None),
    ("AcquisitionTime", 20, None),
    ("DUTID", 20, None),
    ("Operator", 50, None),
    ("DetectorSignature", 20, None),
    ("TestSystemName", 20, None),
    ("TestStationRevision", 20, None),
    ("CoreBundleRevision", 20, None),
    ("AcquisitionName", 40, None),
    ("AcquisitionParameterRevision", 20, None),
    ("OriginalNumberOfRows", 2, "<H"),
    ("OriginalNumberOfColumns", 2, "<H"),
    ("RowNumberUpperLeftPointArchiveROI", 2, "<H"),
    ("ColNumberUpperLeftPointArchiveROI", 2, "<H"),
    ("Swapped", 2, "<H"),
    ("Reordered", 2, "<H"),
    ("HorizontalFlipped", 2, "<H"),
    ("VerticalFlipped", 2, "<H"),
    ("WindowValueDesired", 2, "<H"),
    ("LevelValueDesired", 2, "<H"),
    ("AcquisitionMode", 2, "<H"),
    ("AcquisitionType", 2, "<H"),
    ("UserAcquisitionCoffFileName1", 100, None),
    ("UserAcquisitionCoffFileName2", 100, None),
    ("FramesBeforeExpose", 2, "<H"),
    ("FramesDuringExpose", 2, "<H"),
    ("FramesAfterExpose", 2, "<H"),
    ("IntervalBetweenFrames", 2, "<H"),
    ("ExposeTimeDelayInMicrosecs", 8, "<d"),
    ("TimeBetweenFramesInMicrosecs", 8, "<d"),
    ("FramesToSkipExpose", 2, "<H"),
    ("ExposureMode", 2, "<H"),
    ("PrepPresetTimeInMicrosecs", 8, "<d"),
    ("ExposePresetTimeInMicrosecs", 8, "<d"),
    ("AcquisitionFrameRateInFps", 4, "<f"),
    ("FOVSelect", 2, "<H"),
    ("ExpertMode", 2, "<H"),
    ("SetVCommon1", 8, "<d"),
    ("SetVCommon2", 8, "<d"),
    ("SetAREF", 8, "<d"),
    ("SetAREFTrim", 4, "<L"),
    ("SetSpareVoltageSource", 8, "<d"),
    ("SetCompensationVoltageSource", 8, "<d"),
    ("SetRowOffVoltage", 8, "<d"),
    ("SetRowOnVoltage", 8, "<d"),
    ("StoreCompensationVoltage", 4, "<L"),
    ("RampSelection", 2, "<H"),
    ("TimingMode", 2, "<H"),
    ("Bandwidth", 2, "<H"),
    ("ARCIntegrator", 2, "<H"),
    ("ARCPostIntegrator", 2, "<H"),
    ("NumberOfRows", 4, "<L"),
    ("RowEnable", 2, "<H"),
    ("EnableStretch", 2, "<H"),
    ("CompEnable", 2, "<H"),
    ("CompStretch", 2, "<H"),
    ("LeftEvenTristate", 2, "<H"),
    ("RightOddTristate", 2, "<H"),
    ("TestModeSelect", 4, "<L"),
    ("AnalogTestSource", 4, "<L"),
    ("VCommonSelect", 4, "<L"),
    ("DRCColumnSum", 4, "<L"),
    ("TestPatternFrameDelta", 4, "<L"),
    ("TestPatternRowDelta", 4, "<L"),
    ("TestPatternColumnDelta", 4, "<L"),
    ("DetectorHorizontalFlip", 2, "<H"),
    ("DetectorVerticalFlip", 2, "<H"),
    ("DFNAutoScrubOnOff", 2, "<H"),
    ("FiberChannelTimeOutInMicrosecs", 4, "<L"),
    ("DFNAutoScrubDelayInMicrosecs", 4, "<L"),
    ("StoreAECROI", 2, "<H"),
    ("TestPatternSaturationValue", 2, "<H"),
    ("TestPatternSeed", 4, "<L"),
    ("ExposureTimeInMillisecs", 4, "<f"),
    ("FrameRateInFps", 4, "<f"),
    ("kVp", 4, "<f"),
    ("mA", 4, "<f"),
    ("mAs", 4, "<f"),
    ("FocalSpotInMM", 4, "<f"),
    ("GeneratorType", 20, None),
    ("StrobeIntensityInFtL", 4, "<f"),
    ("NDFilterSelection", 2, "<H"),
    ("RefRegTemp1", 8, "<d"),
    ("RefRegTemp2", 8, "<d"),
    ("RefRegTemp3", 8, "<d"),
    ("Humidity1", 4, "<f"),
    ("Humidity2", 4, "<f"),
    ("DetectorControlTemp", 8, "<d"),
    ("DoseValueInmR", 8, "<d"),
    ("TargetLevelROIRow0", 2, "<H"),
    ("TargetLevelROICol0", 2, "<H"),
    ("TargetLevelROIRow1", 2, "<H"),
    ("TargetLevelROICol1", 2, "<H"),
    ("FrameNumberForTargetLevelROI", 2, "<H"),
    ("PercentRangeForTargetLevel", 2, "<H"),
    ("TargetValue", 2, "<H"),
    ("ComputedMedianValue", 2, "<H"),
    ("LoadZero", 2, "<H"),
    ("MaxLUTOut", 2, "<H"),
    ("MinLUTOut", 2, "<H"),
    ("MaxLinear", 2, "<H"),
    ("Reserved", 2, "<H"),
    ("ElectronsPerCount", 2, "<H"),
    ("ModeGain", 2, "<H"),
    ("TemperatureInDegC", 8, "<d"),
    ("LineRepaired", 2, "<H"),
    ("LineRepairFileName", 100, None),
    ("CurrentLongitudinalInMM", 4, "<f"),
    ("CurrentTransverseInMM", 4, "<f"),
    ("CurrentCircularInMM", 4, "<f"),
    ("CurrentFilterSelection", 4, "<L"),
    ("DisableScrubAck", 2, "<H"),
    ("ScanModeSelect", 2, "<H"),
    ("DetectorAppSwVersion", 20, None),
    ("DetectorNIOSVersion", 20, None),
    ("DetectorPeripheralSetVersion", 20, None),
    ("DetectorPhysicalAddress", 20, None),
    ("PowerDown", 2, "<H"),
    ("InitialVoltageLevel_VCOMMON", 8, "<d"),
    ("FinalVoltageLevel_VCOMMON", 8, "<d"),
    ("DmrCollimatorSpotSize", 10, None),
    ("DmrTrack", 5, None),
    ("DmrFilter", 5, None),
    ("FilterCarousel", 2, "<H"),
    ("Phantom", 20, None),
    ("SetEnableHighTime", 2, "<H"),
    ("SetEnableLowTime", 2, "<H"),
    ("SetCompHighTime", 2, "<H"),
    ("SetCompLowTime", 2, "<H"),
    ("SetSyncLowTime", 2, "<H"),
    ("SetConvertLowTime", 2, "<H"),
    ("SetSyncHighTime", 2, "<H"),
    ("SetEOLTime", 2, "<H"),
    ("SetRampOffsetTime", 2, "<H"),
    ("FOVStartingValue", 2, "<H"),
    ("ColumnBinning", 2, "<H"),
    ("RowBinning", 2, "<H"),
    ("BorderColumns64", 2, "<H"),
    ("BorderRows64", 2, "<H"),
    ("FETOffRows64", 2, "<H"),
    ("FOVStartColumn128", 2, "<H"),
    ("FOVStartRow128", 2, "<H"),
    ("NumberOfColumns128", 2, "<H"),
    ("NumberOfRows128", 2, "<H"),
    ("VFPAquisition", 2000, None),
    ("Comment", 200, None),
]


class GeImage(FabioImage):
    """GE a-Si Angio detector file format."""

    DESCRIPTION = "GE a-Si Angio detector file format"

    DEFAULT_EXTENSIONS = []

    _need_a_seek_to_read = True

    BITDEPTH_TO_DATATYPES = {
        8: numpy.uint8,
        16: numpy.uint16,
        32: numpy.uint32,
    }

    BLANKED_HEADER_METADATA = {
        "StandardHeaderSizeInBytes": 8192,
        "UserHeaderSizeInBytes": 0,
        "NumberOfRowsInFrame": 2048,
        "NumberOfColsInFrame": 2048,
        "ImageDepthInBits": 16,
    }

    BLANK_IMAGE_FORMAT = b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    """At APS they blanked the header with 8192 bytes of zeros when
    updating the GE firmware.  This happened in ~2018 to the best
    of my knowledge, and *a lot* of data has been measured with
    the GE detectors that are missing the magic bytes and header.
    """

    def _readheader(self, infile):
        """Read a GE image header."""

        infile.seek(0)
        self.header = self.check_header()
        for name, nbytes, fmt in GE_HEADER_INFO:
            if fmt is None:
                self.header[name] = infile.read(nbytes)
            else:
                self.header[name] = struct.unpack(fmt, infile.read(nbytes))[0]
        if self.header["ImageFormat"] == self.BLANK_IMAGE_FORMAT:
            # Assume we have a bastardized GE file from
            # Argonne, where they blanked the header with all zeros
            self.header.update(self.BLANKED_HEADER_METADATA)

            # No better way for now to know the number of frames
            if self.header["NumberOfFrames"] == 0:
                cur = infile.tell()
                infile.seek(0, io.SEEK_END)
                file_size = infile.tell()
                infile.seek(cur, io.SEEK_SET)

                depth = self.header["ImageDepthInBits"] // 8
                rows = self.header["NumberOfRowsInFrame"]
                cols = self.header["NumberOfColsInFrame"]
                header_size = self.header["StandardHeaderSizeInBytes"]
                bytes_per_frames = rows * cols * depth
                if not numpy.remainder(file_size, bytes_per_frames) == header_size:
                    raise IOError("GE file size is incorrect")
                nframes = file_size // bytes_per_frames
                self.header["NumberOfFrames"] = nframes
        elif any(
            self.header["ImageFormat"].startswith(x) for x, _ in EDF_MAGIC_NUMBERS
        ):
            # At APS, they started saving some GE headers with an EDF format in ~2022

            # Use the same metadata as the blank header...
            self.header.update(self.BLANKED_HEADER_METADATA)

            # Read the EDF header
            cur = infile.tell()
            infile.seek(0)
            header = EdfImage._read_header_block(infile, 1)
            infile.seek(cur, io.SEEK_SET)

            # Extract any needed info from the EDF header
            self.header["NumberOfFrames"] = int(header[0]["Num_Images"])

    def read(self, fname, frame=None):
        """Read header into self.header and the data into self.data."""
        if frame is None:
            frame = 0
        self.header = self.check_header()
        self.resetvals()
        infile = self._open(fname, "rb")
        self.sequencefilename = fname
        self._readheader(infile)
        self._nframes = self.header["NumberOfFrames"]
        self._readframe(infile, frame)
        infile.close()
        return self

    def _makeframename(self):
        """
        Represent a frame.

        The thing to be printed for the user to represent a frame inside
        a file.
        """
        self.filename = "%s$%04d" % (self.sequencefilename, self.currentframe)

    def _readframe(self, filepointer, img_num):
        """
        Load only one image from the file.

        The first image in the sequence 0; raises an exception if you give an
        invalid image index, otherwise fills in self.data.

        :param filepointer: Pointer to the input file stream
        :param int img_num: Number of the frame (0 = 1st frame)
        :raises: IndexError if the requested frame number is outside of the
            available frames
        :raises: IOError if there is problem to decode or read the frame
        """
        if not (0 <= img_num < self.nframes):
            raise IndexError("Bad image number")

        cols = self.header["NumberOfColsInFrame"]
        rows = self.header["NumberOfRowsInFrame"]
        bitdepth = self.header["ImageDepthInBits"]
        standard_header_size = self.header["StandardHeaderSizeInBytes"]
        user_header_size = self.header["UserHeaderSizeInBytes"]

        imglength = cols * rows * (bitdepth // 8)
        imgstart = standard_header_size + user_header_size + img_num * imglength
        filepointer.seek(imgstart, io.SEEK_SET)

        datatype = self.BITDEPTH_TO_DATATYPES.get(bitdepth, None)
        if datatype is None:
            raise IOError("Data depth format %sbits is not supported" % bitdepth)

        raw = filepointer.read(imglength)
        data = numpy.frombuffer(raw, self.get_stype(datatype, "little")).copy()
        data = data.reshape(rows, cols)
        self.data = data
        self._shape = None
        self.currentframe = int(img_num)
        self._makeframename()

    def getframe(self, num):
        """Return a frame as a new FabioImage object."""
        if not (0 <= num < self.nframes):
            raise IndexError("Requested frame number is out of range")
        # Do a deep copy of the header to make a new one
        newheader = {}
        for k in self.header.keys():
            newheader[k] = self.header[k]
        frame = GeImage(header=newheader)
        # ??? isn't this a single frame by contruction?
        frame._nframes = self.nframes
        frame.sequencefilename = self.sequencefilename
        infile = frame._open(self.sequencefilename, "rb")
        frame._readframe(infile, num)
        infile.close()
        return frame

    def next(self):
        """Get the next image in a series as a fabio image."""
        if self.currentframe < (self.nframes - 1) and self.nframes > 1:
            return self.getframe(self.currentframe + 1)
        else:
            newobj = GeImage()
            newobj.read(next_filename(self.sequencefilename))
            return newobj

    def previous(self):
        """Get the previous image in a series as a fabio image."""
        if self.currentframe > 0:
            return self.getframe(self.currentframe - 1)
        else:
            newobj = GeImage()
            newobj.read(previous_filename(self.sequencefilename))
            return newobj


GEimage = GeImage
